﻿Public Class Chip_8
#Region "Variablen"
    Public inst As UShort          ' Instruction Variable (2Byte)
    Public memory(4096) As Byte    ' 4K Speicher
    Public V(16) As Byte           ' 15 8bit-Register + "Carry Flag-Register"
    Public IR As UShort            ' Index Register
    Public PC As UShort            ' Program Counter von 0x000 bis 0xFFF
    ' **** System Memory Map ****
    ' 0x000-0x1FF - Chip8-Emulator/Interpreter (enthält Zeichentabelle)
    ' 0x050-0x0A0 - Speicherbereich für die 4x5 Pixel Zeichenset (0-F)
    ' 0x200-0xFFF - Programmspeicher und Arbeitsspeicher
    ' **** ----------------- ****
    Public screen(64 * 32) As Byte ' Monochromes Display, 64 x 32 Pixel gross
    Public updatescreen As Boolean = False ' Gibt zurück ob sich der Displayinhalt geändert hat
    Public delay_timer As Byte     ' Delay-Timer mit 60Hz
    Public beep_timer As Byte      ' Sound-Timer für den Buzzer/Beeper
    Public stack(16) As UShort     ' Der "Stack", benötigt wenn man Subroutinen aufrufen will
    Public SP As UShort            ' Der Stack Pointer
    Public KEY(16) As Byte         ' Eingabe des Chip-8(Hex-Keypad von 0x0-0xF)
    Dim chartable() As Byte = _
{
  &HF0, &H90, &H90, &H90, &HF0, _
  &H20, &H60, &H20, &H20, &H70, _
  &HF0, &H10, &HF0, &H80, &HF0, _
  &HF0, &H10, &HF0, &H10, &HF0, _
  &H90, &H90, &HF0, &H10, &H10, _
  &HF0, &H80, &HF0, &H10, &HF0, _
  &HF0, &H80, &HF0, &H90, &HF0, _
  &HF0, &H10, &H20, &H40, &H40, _
  &HF0, &H90, &HF0, &H90, &HF0, _
  &HF0, &H90, &HF0, &H10, &HF0, _
  &HF0, &H90, &HF0, &H90, &H90, _
  &HE0, &H90, &HE0, &H90, &HE0, _
  &HF0, &H80, &H80, &H80, &HF0, _
  &HE0, &H90, &H90, &H90, &HE0, _
  &HF0, &H80, &HF0, &H80, &HF0, _
  &HF0, &H80, &HF0, &H80, &H80 _
}
#End Region

    ''' <summary>
    ''' Initialisiert den Chip8-Emulator
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub init()
        PC = &H200  ' Das Programm startet bei 0x200
        inst = 0    ' Instruction-Variable löschen
        IR = 0      ' Index Register löschen
        SP = 0      ' StackPointer löschen

        For i As UShort = 0 To 4096
            memory(i) = &H0 'Speicher clearen
            If i <= 16 Then
                V(i) = &H0  'Register clearen
                stack(i) = &H0 'Stack clearen
            End If
            If i <= 2048 Then
                screen(i) = &H0 'Displayinhalt löschen
            End If
        Next

        'Timer zurücksetzten
        delay_timer = 0
        beep_timer = 0

        updatescreen = True 'Screen clearen

        ' Zeichentabelle laden
        For i As UShort = 0 To 79
            memory(i) = chartable(i)
        Next
    End Sub

    ''' <summary>
    ''' Arbeitet einen Takt ab
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub emucyc()
        inst = CUShort(memory(PC) * 256 Or memory(PC + 1))
        Debug.Print("0x" & inst.ToString("X"))
        Select Case (inst And &HF000) 'PS: Ich hasse die 'HEX' Schreibweise in VB, einfach nur grässlich...
            Case &H0
                Select Case (inst And &HFFF)
                    Case &HE0 'Displayinhalt löschen
                        For i As UShort = 0 To 2048
                            screen(i) = &H0 'Displayinhalt löschen
                        Next
                        updatescreen = True
                        PC = CUShort(PC + 2)
                    Case &HEE 'Von der Subroutine zurückspringen
                        SP = CUShort(SP - 1)
                        PC = stack(SP)
                        PC = CUShort(PC + 2)
                    Case Else '
                        Console.Write("Unbekannter Befehl")
                End Select
            Case &H1000 'Zur Adresse NNN springen
                PC = CUShort(inst And &HFFF)
            Case &H2000 'Subroutine bei NNN aufrufen
                stack(SP) = PC
                SP = CUShort(SP + 1)
                PC = CUShort(inst And &HFFF)
            Case &H3000 'Nächster Befehl überspringen wenn VX == NN
                If (V((inst And &HF00) >> 8) = (inst And &HFF)) Then
                    PC = CUShort(PC + 4)
                Else
                    PC = CUShort(PC + 2)
                End If
            Case &H4000 'Nächster Befehl überspringen wenn VX != NN
                If (V((inst And &HF00) >> 8) = (inst And &HFF)) Then
                    PC = CUShort(PC + 4)
                Else
                    PC = CUShort(PC + 2)
                End If
            Case &H5000 'Nächster Befehl überspringen wenn VX == VY
                If (V((inst And &HF00) >> 8) = (inst And &HFF)) Then
                    PC = CUShort(PC + 4)
                Else
                    PC = CUShort(PC + 2)
                End If
            Case &H6000 'Setzte VX auf NN
                V((inst And &HF00) >> 8) = CByte(inst And &HFF)
                PC = CUShort(PC + 2)
            Case &H7000 'Addiert NN zu VX
                Dim i As Integer = V((inst And &HF00) >> 8)
                Dim j As Integer = (inst And &HFF)
                V((inst And &HF00) >> 8) = CByte((i + j) And &HFF)
                PC = CUShort(PC + 2)
            Case &H8000
                Select Case (inst And &HF)
                    Case &H0 ' Setzte VX mit dem Wert von VY
                        V((inst And &HF00) >> 8) = V((inst And &HF0) >> 4)
                        PC = CUShort(PC + 2)
                    Case &H1 ' Setzt VX auf "VX oder VY"
                        V((inst And &HF00) >> 8) = V((inst And &HF00) >> 8) Or V((inst And &HF0) >> 4)
                        PC = CUShort(PC + 2)
                    Case &H2 ' Setzt VX auf "VX und VY"
                        V((inst And &HF00) >> 8) = V((inst And &HF00) >> 8) And V((inst And &HF0) >> 4)
                        PC = CUShort(PC + 2)
                    Case &H3 ' Setzt VX auf "VX Exklusiv-Oder VY"
                        V((inst And &HF00) >> 8) = V((inst And &HF00) >> 8) Xor V((inst And &HF0) >> 4)
                        PC = CUShort(PC + 2)
                    Case &H4 ' Addiert VY zu VX. VF ist 1, wenn ein Rest bleibt
                        If V((inst And &HF0) >> 4) > (&HFF - V((inst And &HF00) >> 8)) Then
                            V(&HF) = 1 'VF = 1 bei Rest
                        Else
                            V(&HF) = 0
                        End If
                        Dim i As Integer = V((inst And &HF00) >> 8)
                        Dim j As Integer = V((inst And &HF0) >> 4)

                        V((inst And &HF00) >> 8) = CByte((i+j) And &HFF)
                        PC = CUShort(PC + 2)
                    Case &H5 ' Subtrahiert VY von VX. VF ist 1 wenn Was bleibt
                        If V((inst And &HF0) >> 4) > V((inst And &HF00) >> 8) Then
                            V(&HF) = 1 'VF = 1 bei Rest
                        Else
                            V(&HF) = 0
                        End If
                        Dim i As Integer = V((inst And &HF00) >> 8)
                        Dim j As Integer = V((inst And &HF0) >> 4)
                        V((inst And &HF00) >> 8) = CByte((i - j) And &HFF)
                        'V((inst And &HF00) >> 8) -= V((inst And &HF0) >> 4)
                        PC = CUShort(PC + 2)
                    Case &H6 ' Schiebt VX nach rechts um eine Stelle. VF ist = 0 bevor das letzte Signifikante Bit von VX in VF geschoben wird
                        V(&HF) = CByte(V((inst And &HF00) >> 8) And &H1)
                        V((inst And &HF00) >> 8) = V((inst And &HF00) >> 8) >> 1
                        PC = CUShort(PC + 2)
                    Case &H7 ' Setzt VX in VY minux VX. VF ist = 0 wenns ein Rest gibt, und 1 wenn nicht
                        If (V((inst And &HF00) >> 8) > V((inst And &HF0) >> 4)) Then
                            V(&HF) = 0
                        Else
                            V(&HF) = 1
                        End If
                        V((inst And &HF00) >> 8) = V((inst And &HF0) >> 4)
                        PC = CUShort(PC + 2)
                    Case &HE ' Shiebt VX nach links um eine Stelle. VF übernimmt den Wert des MostSignificantBit von VX bevor es geschoben wird
                        V(&HF) = V((inst And &HF00) >> 8) >> 7
                        V((inst And &HF00) >> 8) = V((inst And &HF00) >> 8) << 1
                        PC = CUShort(PC + 2)
                End Select
            Case &H9000 ' Überspringt den nächsten Befehl wenn VX nicht gleich VY ist
                If Not V((inst And &HF00) >> 8) = V((inst And &HF0) >> 4) Then
                    PC = CUShort(PC + 4)
                Else
                    PC = CUShort(PC + 2)
                End If
            Case &HA000 ' Setzt I mit dem Wert NNN
                IR = CUShort(inst And &HFFF)
                PC = CUShort(PC + 2)
            Case &HB000 ' Springt zur Adresse NNN plus V0
                PC = CUShort((inst And &HFFF) + V(0))
            Case &HC000 ' Setzt VX auf einem zufälligen Wert und NN
                V((inst And &HF00) >> 8) = CByte(CByte(Math.Ceiling(Rnd() * &HFF)) And (inst And &HFF))
                PC = CUShort(PC + 2)
            Case &HD000 ' Zeichnet ein Sprite an der Koordinate (VX, VY), 
                ' mit einer Breite von 8 Pixeln und einer Höhe von N Pixel.
                'Jede Zeile von 8 Pixeln ist als bit-codiert, startend vom Speicherplatz I.
                'I ändert sich nicht nach der Ausführung dieses Befehls.
                'VF ist auf 1 gesetzt ist, wenn irgendwelche Bildschirm-Pixel getoggelt werden
                Dim x As UShort = V((inst And &HF00) >> 8)
                Dim y As UShort = V((inst And &HF0) >> 4)
                Dim höhe As UShort = CUShort(inst And &HF)
                Debug.Print(x.ToString)
                Debug.Print(y.ToString)
                Debug.Print(höhe.ToString)
                Dim pixel As UShort
                V(&HF) = 0
                For yline As Integer = 0 To (höhe - 1)
                    pixel = memory(IR + yline)
                    For xline As Integer = 0 To 7
                        If (Not CDbl(pixel & (&H80 >> xline)) = 0) Then
                            If Not (pixel And (&H80 >> xline)) = 0 Then
                                If (screen((x + xline + ((y + yline) * 64)) And &H7FF) = 1) Then
                                    V(&HF) = 1
                                End If
                                screen((x + xline + ((y + yline) * 64)) And &H7FF) = CByte(screen((x + xline + ((y + yline) * 64)) And &H7FF) Xor 1)
                            End If
                        End If
                    Next
                Next
                updatescreen = True
                PC = CUShort(PC + 2)
            Case &HE000
                Select Case (inst And &HFF)
                    Case &H9E 'Überspringt den nächsten Befehl wenn der Schalter in VX gedrückt ist
                        If (Not (KEY(V((inst And &HF00) >> 8)) = 0)) Then
                            PC = CUShort(PC + 4)
                        Else
                            PC = CUShort(PC + 2)
                        End If
                    Case &HA1 'Überspringt den nächsten Befehl wenn der Schalter in VX nicht gedrückt ist
                        If (KEY(V((inst And &HF00) >> 8)) = 0) Then
                            PC = CUShort(PC + 4)
                        Else
                            PC = CUShort(PC + 2)
                        End If
                End Select
            Case &HF000
                Select Case inst And &HFF
                    Case &H7 ' Setzt VX mit dem Wert des Delay-Timers
                        V((inst And &HF00) >> 8) = delay_timer
                        PC = CUShort(PC + 2)
                    Case &HA ' Wartet ein Tastendruck ab und legt ihn in VX ab
                        Dim Taste_gedrückt As Boolean = False
                        For i As Integer = 0 To 15
                            If Not KEY(i) = 0 Then
                                V((inst And &HF00) >> 8) = CByte(i)
                                Taste_gedrückt = True
                            End If
                        Next

                        If Not Taste_gedrückt Then Return
                        PC = CUShort(PC + 2)
                    Case &H15 ' Setzt den Delay-Timer mit dem Wert von VX
                        delay_timer = V((inst And &HF00) >> 8)
                        PC = CUShort(PC + 2)
                    Case &H18 ' Setzt den Beep-Timer mit dem Wert von VX
                        beep_timer = V((inst And &HF00) >> 8)
                        PC = CUShort(PC + 2)
                    Case &H1E ' Addiert VX zu IR
                        If (IR + V((inst And &HF00) >> 8) > &HFFF) Then
                            V(&HF) = 1
                        Else
                            V(&HF) = 0
                        End If
                        IR += V((inst And &HF00) >> 8)
                        PC = CUShort(PC + 2)
                    Case &H29 ' Setzt IR zur position des Sprites für das Zeichen in VX. Zeichen 0-F(in Hex) sind in einer 4x5-Zeichentabelle vorhanden
                        IR = CUShort(V((inst And &HF00) >> 8) * &H5)
                        PC = CUShort(PC + 2)
                    Case &H33 ' 
                        memory(IR) = CByte(V((inst And &HF00) >> 8) / 100)
                        memory(IR + 1) = CByte((V((inst And &HF00) >> 8) / 10) Mod 10)
                        memory(IR + 2) = CByte((V((inst And &HF00) >> 8) Mod 100) Mod 10)
                        PC = CUShort(PC + 2)
                    Case &H55 ' Speichert V0 in VX in dem Speicher IR
                        For i As Integer = 0 To ((inst And &HF00) >> 8)
                            memory(IR + i) = V(i)
                        Next
                        IR = CUShort(IR + (((inst And &HF00) >> 8) + 1))
                        PC = CUShort(PC + 2)
                    Case &H65 ' Füllt V0 zu VX mit den Werten des Speichers bei der Adresse IR
                        For i As Integer = 0 To ((inst And &HF00) >> 8)
                            V(i) = memory(IR + i)
                        Next
                        IR = CUShort(IR + (((inst And &HF00) >> 8) + 1))
                        PC = CUShort(PC + 2)
                End Select
        End Select
        If delay_timer > 0 Then delay_timer = CByte(delay_timer - 1)
        If beep_timer > 0 Then
            If beep_timer = 1 Then
                Console.Beep()
            End If
            beep_timer = CByte(beep_timer - 1)
        End If
    End Sub

    ''' <summary>
    ''' Ladet ein Chip-8 Spiel in den Speicher
    ''' </summary>
    ''' <param name="Pfad">Pfad des Spiels</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function LoadChip8(ByVal Pfad As String) As Integer
        init()
        If Pfad.Trim = Nothing Then
            Return 1    'Kein Pfad 
        End If
        Dim reader As New IO.BinaryReader(IO.File.Open(Pfad, IO.FileMode.Open))
        If reader.BaseStream.Length = 0 Or reader.BaseStream.Length >= (4096 - 512) Then
            Debug.Print(reader.BaseStream.Length.ToString)
            Return 2    'Dateigrösse ungültig
        End If
        Dim data As Byte() = reader.ReadBytes(CInt(reader.BaseStream.Length))
        For i As Integer = 0 To CInt((reader.BaseStream.Length - 1))
            memory(512 + i) = data(i) 'Daten in den Speicher einlesen
        Next

        Return 0
    End Function

    Public Function GetRandom(ByVal Min As Integer, ByVal Max As Integer) As Integer
        Dim Generator As System.Random = New System.Random()
        Return Generator.Next(Min, Max)
    End Function
End Class